{% extends "layout.html" %}
{% set title = 'scikit-learn: machine learning in Python' %}

{% if theme_link_to_live_contributing_page|tobool %}
  {%- set contributing_link = "https://scikit-learn.org/dev/developers/contributing.html" %}
  {%- set contributing_attrs = 'target="_blank" rel="noopener noreferrer"' %}
{%- else %}
  {%- set contributing_link = pathto('developers/contributing') %}
  {%- set contributing_attrs = '' %}
{%- endif %}


{% block content %}
<div class="container-fluid sk-landing-bg py-3">
  <div class="container sk-landing-container">
    <div class="row">
      <div class="col-md-6 mb-3 mb-md-0">
        <h1 class="sk-landing-header text-white text-monospace">scikit-learn</h1>
        <h4 class="sk-landing-subheader text-white font-italic mb-3">Machine Learning in Python</h4>
        <a class="btn sk-landing-btn mb-1" href="{{ pathto('getting_started') }}" role="button">Getting Started</a>
        <a class="btn sk-landing-btn mb-1" href="{{ pathto(release_highlights) }}" role="button">Release Highlights for {{ release_highlights_version }}</a>
        <a class="btn sk-landing-btn mb-1" href="https://github.com/scikit-learn/scikit-learn" role="button">GitHub</a>
      </div>
      <div class="col-md-6 d-flex">
        <ul class="sk-landing-header-body">
          <li>Simple and efficient tools for predictive data analysis</li>
          <li>Accessible to everybody, and reusable in various contexts</li>
          <li>Built on NumPy, SciPy, and matplotlib</li>
          <li>Open source, commercially usable - BSD license</li>
        </ul>
      </div>
    </div>
  </div>
</div>

<div class="container sk-landing-container pt-3 body" role="main">
  <div class="row no-gutters">
    <div class="col-md-4 mb-3 px-md-2 sk-px-xl-4">
      <div class="card h-100">
        <div class="card-body">
          <a href="supervised_learning.html#supervised-learning"><h4 class="sk-card-title card-title">Classification</h4></a>
          <p class="card-text">Identifying which category an object belongs to.</p>
          <p class="card-text"><strong>Applications:</strong> Spam detection, image recognition.</br>
          <strong>Algorithms:</strong>
          <a href="modules/svm.html#svm-classification">SVM</a>,
          <a href="modules/neighbors.html#classification">nearest neighbors</a>,
          <a href="modules/ensemble.html#forest">random forest</a>,
          and <a href="supervised_learning.html#supervised-learning">more...</a></p>
        </div>
        <div class="overflow-hidden mx-2 text-center flex-fill">
          <a href="auto_examples/classification/plot_classifier_comparison.html"  aria-label="Classification">
          <img src="_images/sphx_glr_plot_classifier_comparison_001_carousel.png" class="sk-index-img" style="width:initial;max-width:initial" alt="Classifier comparison">
          </a>
        </div>
          <a href="auto_examples/index.html#classification" class="sk-btn-primary btn text-white btn-block" role="button">Examples</a>
      </div>
    </div>
    <div class="col-md-4 mb-3 px-md-2 sk-px-xl-4">
      <div class="card h-100">
        <div class="card-body">
          <a href="supervised_learning.html#supervised-learning"><h4 class="sk-card-title card-title">Regression</h4></a>
          <p class="card-text">Predicting a continuous-valued attribute associated with an object.</p>
          <p class="card-text"><strong>Applications:</strong> Drug response, Stock prices.</br>
          <strong>Algorithms:</strong>
          <a href="modules/svm.html#svm-regression">SVR</a>,
          <a href="modules/neighbors.html#regression">nearest neighbors</a>,
          <a href="modules/ensemble.html#forest">random forest</a>,
          and <a href="supervised_learning.html#supervised-learning">more...</a></p>
        </div>
        <div class="overflow-hidden mx-2 text-center flex-fill">
          <a href="auto_examples/ensemble/plot_adaboost_regression.html"  aria-label="Regression">
          <img src="_images/sphx_glr_plot_adaboost_regression_thumb.png" class="sk-index-img" alt="Decision Tree Regression with AdaBoost">
          </a>
        </div>
          <a href="auto_examples/index.html#examples" class="sk-btn-primary btn text-white btn-block" role="button">Examples</a>
      </div>
    </div>
    <div class="col-md-4 mb-3 px-md-2 sk-px-xl-4">
      <div class="card h-100">
        <div class="card-body">
          <a href="modules/clustering.html#clustering"><h4 class="sk-card-title card-title">Clustering</h4></a>
          <p class="card-text">Automatic grouping of similar objects into sets.</p>
          <p class="card-text"><strong>Applications:</strong> Customer segmentation, Grouping experiment outcomes</br>
          <strong>Algorithms:</strong>
          <a href="modules/clustering.html#k-means">k-Means</a>,
          <a href="modules/clustering.html#spectral-clustering">spectral clustering</a>,
          <a href="modules/clustering.html#mean-shift">mean-shift</a>,
          and <a href="modules/clustering.html#clustering">more...</a></p>
        </div>
        <div class="overflow-hidden mx-2 text-center flex-fill">
          <a href="auto_examples/cluster/plot_kmeans_digits.html"  aria-label="Clustering">
          <img src="_images/sphx_glr_plot_kmeans_digits_thumb.png" class="sk-index-img" alt="A demo of K-Means clustering on the handwritten digits data">
          </a>
        </div>
          <a href="auto_examples/index.html#cluster-examples" class="sk-btn-primary btn text-white btn-block" role="button">Examples</a>
      </div>
    </div>
    <div class="col-md-4 mb-3 px-md-2 sk-px-xl-4">
      <div class="card h-100">
        <div class="card-body">
          <a href="modules/decomposition.html#decompositions"><h4 class="sk-card-title card-title">Dimensionality reduction</h4></a>
          <p class="card-text">Reducing the number of random variables to consider.</p>
          <p class="card-text"><strong>Applications:</strong> Visualization, Increased efficiency</br>
          <strong>Algorithms:</strong>
          <a href="modules/decomposition.html#pca">PCA</a>,
          <a href="modules/feature_selection.html#feature-selection">feature selection</a>,
          <a href="modules/decomposition.html#nmf">non-negative matrix factorization</a>,
          and <a href="modules/decomposition.html#decompositions">more...</a></p>
        </div>
        <div class="overflow-hidden mx-2 text-center flex-fill">
          <a href="auto_examples/decomposition/plot_pca_iris.html"  aria-label="Dimensionality reduction">
          <img src="_images/sphx_glr_plot_pca_iris_thumb.png" class="sk-index-img" alt="PCA example with Iris Data-set">
          </a>
        </div>
          <a href="auto_examples/index.html#decomposition-examples" class="sk-btn-primary btn text-white btn-block" role="button">Examples</a>
      </div>
    </div>
    <div class="col-md-4 mb-3 px-md-2 sk-px-xl-4">
      <div class="card h-100">
        <div class="card-body">
          <a href="model_selection.html#model-selection"><h4 class="sk-card-title card-title">Model selection</h4></a>
          <p class="card-text">Comparing, validating and choosing parameters and models.</p>
          <p class="card-text"><strong>Applications:</strong> Improved accuracy via parameter tuning</br>
          <strong>Algorithms:</strong>
          <a href="modules/grid_search.html#grid-search">grid search</a>,
          <a href="modules/cross_validation.html#cross-validation">cross validation</a>,
          <a href="modules/model_evaluation.html#model-evaluation">metrics</a>,
          and <a href="model_selection.html">more...</a></p>
        </div>
        <div class="overflow-hidden mx-2 text-center flex-fill">
          <a href="auto_examples/model_selection/plot_multi_metric_evaluation.html"  aria-label="Model selection">
            <img src="_images/sphx_glr_plot_multi_metric_evaluation_thumb.png" class="sk-index-img" alt="Demonstration of multi-metric evaluation on cross_val_score and GridSearchCV">
          </a>
        </div>
          <a href="auto_examples/index.html#model-selection" class="sk-btn-primary btn text-white btn-block" role="button">Examples</a>
      </div>
    </div>
    <div class="col-md-4 mb-3 px-md-2 sk-px-xl-4">
      <div class="card h-100">
        <div class="card-body">
          <a href="modules/preprocessing.html#preprocessing"><h4 class="sk-card-title card-title">Preprocessing</h4></a>
          <p class="card-text">Feature extraction and normalization.</p>
          <p class="card-text"><strong>Applications:</strong>  Transforming input data such as text for use with machine learning algorithms.</br>
          <strong>Algorithms:</strong>
          <a href="modules/preprocessing.html#preprocessing">preprocessing</a>,
          <a href="modules/feature_extraction.html#feature-extraction">feature extraction</a>,
          and <a href="modules/preprocessing.html#preprocessing">more...</a></p>
        </div>
        <div class="overflow-hidden mx-2 text-center flex-fill">
          <a href="auto_examples/preprocessing/plot_discretization_strategies.html"  aria-label="Preprocessing">
          <img src="_images/sphx_glr_plot_discretization_strategies_thumb.png" class="sk-index-img" alt="Demonstrating the different strategies of KBinsDiscretizer">
          </a>
        </div>
          <a href="auto_examples/index.html#preprocessing" class="sk-btn-primary btn text-white btn-block" role="button">Examples</a>
      </div>
    </div>
  </div>
</div>

<div class="container-fluid sk-landing-bg-more-info py-3">
  <div class="container sk-landing-container">
    <div class="row">
      <div class="col-md-4">
        <h4 class="sk-landing-call-header">News</h4>
        <ul class="sk-landing-call-list list-unstyled">
        <li><strong>On-going development:</strong>
        <a href="https://scikit-learn.org/dev/whats_new.html"><strong>What's new</strong> (Changelog)</a>
        </li>
        <li><strong>March 2023.</strong> scikit-learn 1.2.2 is available for download (<a href="whats_new/v1.2.html#version-1-2-2">Changelog</a>).
        </li>
        <li><strong>January 2023.</strong> scikit-learn 1.2.1 is available for download (<a href="whats_new/v1.2.html#version-1-2-1">Changelog</a>).
        </li>
        <li><strong>December 2022.</strong> scikit-learn 1.2.0 is available for download (<a href="whats_new/v1.2.html#version-1-2-0">Changelog</a>).
        </li>
        <li><strong>October 2022.</strong> scikit-learn 1.1.3 is available for download (<a href="whats_new/v1.1.html#version-1-1-3">Changelog</a>).
        </li>
        <li><strong>August 2022.</strong> scikit-learn 1.1.2 is available for download (<a href="whats_new/v1.1.html#version-1-1-2">Changelog</a>).
        </li>
        <li><strong>May 2022.</strong> scikit-learn 1.1.1 is available for download (<a href="whats_new/v1.1.html#version-1-1-1">Changelog</a>).
        </li>
        <li><strong>May 2022.</strong> scikit-learn 1.1.0 is available for download (<a href="whats_new/v1.1.html#version-1-1-0">Changelog</a>).
        </li>
        <li><strong>December 2021.</strong> scikit-learn 1.0.2 is available for download (<a href="whats_new/v1.0.html#version-1-0-2">Changelog</a>).
        </li>
        <li><strong>October 2021.</strong> scikit-learn 1.0.1 is available for download (<a href="whats_new/v1.0.html#version-1-0-1">Changelog</a>).
        </li>
        <li><strong>September 2021.</strong> scikit-learn 1.0 is available for download (<a href="whats_new/v1.0.html#version-1-0">Changelog</a>).
        </li>
        <li><strong>April 2021.</strong> scikit-learn 0.24.2 is available for download (<a href="whats_new/v0.24.html#version-0-24-2">Changelog</a>).
        </li>
        <li><strong>January 2021.</strong> scikit-learn 0.24.1 is available for download (<a href="whats_new/v0.24.html#version-0-24-1">Changelog</a>).
        </li>
        <li><strong>December 2020.</strong> scikit-learn 0.24.0 is available for download (<a href="whats_new/v0.24.html#version-0-24-0">Changelog</a>).
        </li>
        <li><strong>August 2020.</strong> scikit-learn 0.23.2 is available for download (<a href="whats_new/v0.23.html#version-0-23-2">Changelog</a>).
        </li>
        <li><strong>May 2020.</strong> scikit-learn 0.23.1 is available for download (<a href="whats_new/v0.23.html#version-0-23-1">Changelog</a>).
        </li>
        <li><strong>May 2020.</strong> scikit-learn 0.23.0 is available for download (<a href="whats_new/v0.23.html#version-0-23-0">Changelog</a>).
        </li>
        <li><strong>Scikit-learn from 0.23 requires Python 3.6 or newer.</strong>
        </li>
        <li><strong>March 2020.</strong> scikit-learn 0.22.2 is available for download (<a href="whats_new/v0.22.html#version-0-22-2">Changelog</a>).
        <li><strong>January 2020.</strong> scikit-learn 0.22.1 is available for download (<a href="whats_new/v0.22.html#version-0-22-1">Changelog</a>).
        <li><strong>December 2019.</strong> scikit-learn 0.22 is available for download (<a href="whats_new/v0.22.html#version-0-22-0">Changelog</a> and <a href="{{ pathto('auto_examples/release_highlights/plot_release_highlights_0_22_0') }}">Release Highlights</a>).
        </li>
        </ul>
      </div>
      <div class="col-md-4">
        <h4 class="sk-landing-call-header">Community</h4>
        <ul class="sk-landing-call-list list-unstyled">
        <li><strong>About us:</strong> See <a href="about.html#people">authors</a> and <a href="{{ contributing_link }}" {{ contributing_attrs }}>contributing</a></li>
        <li><strong>More Machine Learning:</strong> Find <a href="related_projects.html">related projects</a></li>
        <li><strong>Questions?</strong> See <a href="faq.html">FAQ</a> and <a href="https://stackoverflow.com/questions/tagged/scikit-learn">stackoverflow</a></li>
        <li><strong>Subscribe to the</strong> <a href="https://mail.python.org/mailman/listinfo/scikit-learn">mailing list</a></li>
        <li><strong>Gitter:</strong> <a href="https://gitter.im/scikit-learn/scikit-learn">gitter.im/scikit-learn</a></li>
        <li><strong>Blog:</strong> <a href="https://blog.scikit-learn.org">blog.scikit-learn.org</a></li>
        <li><strong>Logos & Branding:</strong> <a href="https://github.com/scikit-learn/scikit-learn/tree/main/doc/logos">logos and branding</a></li>
        <li><strong>Calendar:</strong> <a href="https://blog.scikit-learn.org/calendar/">calendar</a></li>
        <li><strong>Twitter:</strong> <a href="https://twitter.com/scikit_learn">@scikit_learn</a></li>
        <li><strong>Twitter (commits):</strong> <a href="https://twitter.com/sklearn_commits">@sklearn_commits</a></li>
        <li><strong>LinkedIn:</strong> <a href="https://www.linkedin.com/company/scikit-learn">linkedin/scikit-learn</a></li>
        <li><strong>YouTube:</strong> <a href="https://www.youtube.com/channel/UCJosFjYm0ZYVUARxuOZqnnw/playlists">youtube.com/scikit-learn</a></li>
        <li><strong>Facebook:</strong> <a href="https://www.facebook.com/scikitlearnofficial/">@scikitlearnofficial</a></li>
        <li><strong>Instagram:</strong> <a href="https://www.instagram.com/scikitlearnofficial/">@scikitlearnofficial</a></li>
        <li><strong>TikTok:</strong> <a href="https://www.tiktok.com/@scikit.learn">@scikit.learn</a></li>
        <li>Communication on all channels should respect <a href="https://www.python.org/psf/conduct/">PSF's code of conduct.</a></li>
        </ul>

        <a class="btn btn-warning btn-big sk-donate-btn mb-1" href="https://numfocus.org/donate-to-scikit-learn">Help us, <strong>donate!</strong></a>
        <a class="btn btn-warning btn-big mb-1" href="about.html#citing-scikit-learn"><strong>Cite us!</strong></a>
      </div>
      <div class="col-md-4">
        <h4 class="sk-landing-call-header">Who uses scikit-learn?</h4>
        <div id="carouselExampleSlidesOnly" class="carousel slide" data-ride="carousel">
        <div class="carousel-inner">
            <div class="carousel-item active">
            <img class="d-block mx-auto sk-who-uses-carousel-img img-thumbnail" src="_images/inria.png" alt="inria">
            <em>"We use scikit-learn to support leading-edge basic research [...]"</em>
            </div>
            <div class="carousel-item">
            <img class="d-block mx-auto sk-who-uses-carousel-img img-thumbnail" src="_images/spotify.png" alt="spotify">
            <em>"I think it's the most well-designed ML package I've seen so far."</em>
            </div>
            <div class="carousel-item">
            <img class="d-block mx-auto sk-who-uses-carousel-img img-thumbnail" src="_images/change-logo.png" alt="change-logo">
            <em>"scikit-learn's ease-of-use, performance and overall variety of algorithms implemented has proved invaluable [...]."</em>
            </div>
            <div class="carousel-item">
            <img class="d-block mx-auto sk-who-uses-carousel-img img-thumbnail" src="_images/telecomparistech.jpg" alt="telecomparistech">
            <em>"The great benefit of scikit-learn is its fast learning curve [...]"</em>
            </div>
            <div class="carousel-item">
            <img class="d-block mx-auto sk-who-uses-carousel-img img-thumbnail" src="_images/aweber.png" alt="aweber">
            <em>"It allows us to do AWesome stuff we would not otherwise accomplish"</em>
            </div>
            <div class="carousel-item">
            <img class="d-block mx-auto sk-who-uses-carousel-img img-thumbnail" src="_images/yhat.png" alt="yhat">
            <em>"scikit-learn makes doing advanced analysis in Python accessible to anyone."</em>
            </div>
          </div>
        </div>
        <p class="text-right">
            <a href="testimonials/testimonials.html">More testimonials</a>
        </p>
      </div>
    </div>
  </div>
</div>
<div class="container-fluid py-3">
  <div class="container sk-landing-container">
        <a class="sk-footer-funding-link" href="about.html#funding">
        <div class="text-center">
                <p class="mt-2">
                  scikit-learn development and maintenance are financially supported by
                </p>
                <img class="sk-footer-funding-logo" src="_static/inria-small.png" title="INRIA">
                <img class="sk-footer-funding-logo" src="_static/bcg-small.png" title="Boston Consulting Group" >
                <img class="sk-footer-funding-logo" src="_static/axa-small.png" title="AXA Assurances" >
                <img class="sk-footer-funding-logo" src="_static/bnp-small.png" title="BNP Paris Bas Cardif" >
                <img class="sk-footer-funding-logo" src="_static/fujitsu-small.png" title="Fujitsu" >
                <img class="sk-footer-funding-logo" src="_static/microsoft-small.png" title="Microsoft" >
                <img class="sk-footer-funding-logo" src="_static/dataiku-small.png" title="Dataiku" >
                <img class="sk-footer-funding-logo" src="_static/logo_APHP.png" title="APHP" >
                <img class="sk-footer-funding-logo" src="_static/huggingface_logo-noborder.png" title="Hugging Face" >
                <img class="sk-footer-funding-logo" src="_static/quansight-labs-small.png" title="Quansight Labs" >
        </div>
        </a>
  </div>
</div>
{% endblock %}
