Legend for changelogs
---------------------

- |MajorFeature|: something big that you couldn't do before.
- |Feature|: something that you couldn't do before.
- |Efficiency|: an existing feature now may not require as much computation or
  memory.
- |Enhancement|: a miscellaneous minor improvement.
- |Fix|: something that previously didn't work as documentated -- or according
  to reasonable expectations -- should now work.
- |API|: you will need to change your code to have the same effect in the
  future; or a feature will be removed in the future.
